"""The main application code for astare."""

##############################################################################
# Python imports.
import argparse
from typing             import Any
from pathlib            import Path
from importlib.metadata import version

##############################################################################
# Pygments imports.
from pygments.styles import get_all_styles

##############################################################################
# Textual imports.
from textual.app import App

##############################################################################
# Local imports.
from textual_astview import Source, __version__
from .screens        import MainDisplay

##############################################################################
class Astare( App[ None ] ):
    """Main Textual application class."""

    TITLE = "astare"
    """str: The main title of the app."""

    SUB_TITLE = f"A Python AST Explorer ({__version__})"
    """str: The sub title of the app."""

    def __init__( self, cli_args: argparse.Namespace, *args: Any, **kwargs: Any ) -> None:
        """Initialise the app."""
        super().__init__( *args, **kwargs )
        self._args = cli_args

    def on_mount( self ) -> None:
        """Set up the application on startup."""
        self.push_screen( MainDisplay( self._args ) )

##############################################################################
def py_file( path: str ) -> Path:
    """Check that the file we're being asked to look at seems fine

    Args:
        path (str): The argument.

    Returns:
        Path: The `Path` to the file if it looks okay.
    """

    candidate = Path( path )

    if not candidate.exists():
        raise argparse.ArgumentTypeError( f"{path} does not exist" )

    return Path( path )

##############################################################################
def get_args() -> argparse.Namespace:
    """Get the command line arguments.

    Returns:
        argparse.Namespace: The arguments.
    """

    # Create the argument parser object.
    parser = argparse.ArgumentParser(
        prog        = "astare",
        description = "A simple terminal-based AST explorer.",
        epilog      = f"v{__version__}"
    )

    # Add --dark-theme
    parser.add_argument(
        "-t", "--dark-theme",
        choices = list( get_all_styles() ),
        metavar = "THEME",
        default = Source.DEFAULT_DARK_THEME,
        help    = "Set the dark mode theme for the source display.",
    )

    # Add --light-theme
    parser.add_argument(
        "-T", "--light-theme",
        choices = list( get_all_styles() ),
        metavar = "THEME",
        default = Source.DEFAULT_LIGHT_THEME,
        help    = "Set the light mode theme for the source display.",
    )

    # Add --themes
    parser.add_argument(
        "--themes",
        action  = "version",
        help    = "List the available theme names.",
        version = ", ".join( sorted( list( get_all_styles() ) ) )
    )

    # Add --version
    parser.add_argument(
        "-v", "--version",
        help    = "Show version information.",
        action  = "version",
        version = f"%(prog)s {__version__} (Textual v{version( 'textual' )})"
    )

    # The reminder is the file to explore.
    parser.add_argument(
        "file",
        help    = "The file to explore",
        nargs   = "?",
        type    = py_file,
        default = "."
    )

    # Return the arguments.
    return parser.parse_args()

##############################################################################
def main() -> None:
    """Main entry point for astare."""
    Astare( get_args() ).run()

### astare.py ends here



"""An AST viewing widget library plus app, built for and with Textual."""

######################################################################
# Main app information.
__author__     = "Dave Pearson"
__copyright__  = "Copyright 2022, Dave Pearson"
__credits__    = [ "Dave Pearson" ]
__maintainer__ = "Dave Pearson"
__email__      = "davep@davep.org"
__version__    = "0.4.0"
__licence__    = "MIT"

##############################################################################
# Import the widgets to make them easier to get at.
from .widgets.astview     import ASTView, ASTNode
from .widgets.node_info   import NodeInfo
from .widgets.source_info import SourceInfo
from .widgets.source      import Source

##############################################################################
# Export the widgets.
__all__ = [
    "ASTView",
    "ASTNode",
    "NodeInfo",
    "SourceInfo",
    "Source"
]

### __init__.py ends here



"""Main screen for astare."""

##############################################################################
# Python imports.
from typing    import Final, Any, cast, Optional
from argparse  import Namespace
from functools import partial
from pathlib   import Path

##############################################################################
# Textual imports.
from textual.app        import ComposeResult
from textual.binding    import Binding
from textual.screen     import Screen
from textual.reactive   import reactive
from textual.timer      import Timer
from textual.widgets    import Header, Footer, Tree, DirectoryTree
from textual.containers import Horizontal, Vertical
from textual.css.query  import NoMatches

##############################################################################
# Local imports.
from textual_astview import ASTView, ASTNode, NodeInfo, Source, __version__

##############################################################################
class MainDisplay( Screen ):
    """The main display of the app."""

    DEFAULT_CSS = """
    DirectoryTree {
        border: solid $primary-background-lighten-2;
        display: none;
    }

    DirectoryTree.visible {
        display: block;
    }

    DirectiryTree.initial {
        width: 100%;
    }

    DirectoryTree.insert {
        width: 20%;
    }

    DirectoryTree:focus-within {
        border: double $primary-lighten-2;
    }

    Vertical {
        width: 1fr;
    }

    ASTView {
        border: solid $primary-background-lighten-2;
    }

    ASTView:focus-within {
        border: double $primary-lighten-2;
    }
    """

    BINDINGS = [
        Binding( "ctrl+o", "open_new",       "Open" ),
        Binding( "ctrl+r", "toggle_rainbow", "Rainbow" ),
        Binding( "ctrl+d", "toggle_dark",    "Light/Dark" ),
        Binding( "ctrl+q", "app.quit",       "Quit" )
    ]
    """list[ Bindings ]: The bindings for the main screen."""

    UPDATE_DELAY: Final = 0.2
    """float: How long to leave it before updating the highlight in the source."""

    rainbow = reactive( False, init=False )
    """bool: Should 'rainbow' highlighting be used for the source?"""

    def __init__( self, cli_args: Namespace, *args: Any, **kwargs: Any ) -> None:
        """Initialise the main screen.

        Args:
            cli_args (Namespace): The command line arguments.
        """
        super().__init__( *args, **kwargs )
        self._args                       = cli_args
        self._refresh: Optional[ Timer ] = None

    def ast_pane( self ) -> ASTView:
        """Make an ASTView pane.

        Returns:
            ASTView: An ASTView pane for viewing the AST.
        """
        return ASTView( self._args.file )

    def source_pane( self ) -> Source:
        """Make a Source pane.

        Returns:
            Source: A source pane for viewing the code.
        """
        return Source(
            self._args.file,
            dark_theme  = self._args.dark_theme,
            light_theme = self._args.light_theme
        )

    def compose( self ) -> ComposeResult:
        """Compose the main app screen.

        Returns:
            ComposeResult: The result of composing the screen.
        """

        if self._args.file.is_file():
            body = [
                DirectoryTree( str( self._args.file.parent ), classes="insert" ), self.ast_pane(), self.source_pane()
            ]
        else:
            body = [ DirectoryTree( str( self._args.file ), classes="visible initial" ) ]

        yield Header()
        yield Vertical( Horizontal( *body ), NodeInfo() )
        yield Footer()

    def _init_tree( self ) -> None:
        """Set up the tree view when (re)loaded."""
        view = self.query_one( ASTView )
        view.root.expand()
        # Sadly Textual's TreeNode doesn't allow easy access to the
        # children, at the moment, so we're going to do a sneaky dip into
        # the internals here. What I want is to expand up the first couple
        # of levels in the tree, on startup.
        #
        # pylint:disable=protected-access
        if view.root._children:
            view.root._children[ 0 ].expand()
            if view.root._children[ 0 ]._children:
                view.root._children[ 0 ]._children[ 0 ].expand()
        view.focus()

    def on_mount( self ) -> None:
        """Sort the screen once the DOM is mounted."""
        if self.query( ASTView ):
            self._init_tree()
        else:
            self.query_one( DirectoryTree ).focus()

    def highlight_node( self, node: ASTNode ) -> None:
        """Update the display to highlight the given node.

        Args:
            node (ASTNode): The node to highlight.
        """
        self.query_one( NodeInfo ).show( node )
        self.query_one( Source ).highlight( node, self.rainbow )

    async def on_astview_node_highlighted( self, event: ASTView.NodeHighlighted ) -> None:
        """React to a node in the tree being highlighted.

        Args:
            event (ASTTree.NodeHighlighted): The event to react to.
        """
        # If there's a refresh pending...
        if self._refresh is not None:
            # ...stop it.
            await self._refresh.stop()
        # Book in a fresh refresh the full delay on from this event.
        self._refresh = self.set_timer(
            self.UPDATE_DELAY, partial( self.highlight_node, event.node )
        )

    async def on_tree_node_selected( self, event: Tree.NodeSelected[ Any ] ) -> None:
        """React to a node in the tree being selected.

        Args:
            event (Tree.NodeSelected[ Any ]): The event to react to.
        """
        # If there's a refresh pending...
        if self._refresh is not None:
            # ...stop it and nuke all evidence of it.
            await self._refresh.stop()
            self._refresh = None
        # Selected is a mouse click or someone bouncing on the enter key,
        # there's no point in delaying; so let's make it look a bit more
        # snappy by updating right away.
        self.highlight_node( event.node )

    def watch_rainbow( self, new_value: bool ) -> None:
        """React to the rainbow flag being changed.

        Args:
            new_value (bool): The new value for the flag.
        """
        try:
            self.query_one( Source ).highlight(
                cast( ASTNode, self.query_one( ASTView ).cursor_node ), new_value
            )
        except NoMatches:
            pass

    def action_toggle_rainbow( self ) -> None:
        """Toggle the rainbow highlight flag."""
        self.rainbow = not self.rainbow

    def action_toggle_dark( self ) -> None:
        """Toggle daark mode."""
        self.app.dark = not self.app.dark
        try:
            self.query_one( Source ).dark = self.app.dark
            self.highlight_node( cast( ASTNode, self.query_one( ASTView ).cursor_node ) )
        except NoMatches:
            pass

    def action_open_new( self ) -> None:
        """Open a new file for viewing."""
        opener = self.query_one( DirectoryTree )
        opener.toggle_class( "visible" )
        if "visible" in opener.classes:
            opener.focus()

    async def on_directory_tree_file_selected( self, event: DirectoryTree.FileSelected ) -> None:
        """React to a file being selected in the directory tree.

        Args:
            event (DirectoryTree.FileSelected): The file selection event.
        """
        self._args.file = Path( event.path )
        if self.query( ASTView ):
            await self.query_one( ASTView ).remove()
            await self.mount( self.ast_pane(), before="Source" )
        else:
            await self.mount( self.ast_pane(), after="DirectoryTree" )
            await self.mount( self.source_pane(), after="ASTView" )
            self.query_one( DirectoryTree ).toggle_class( "initial", "insert" )
        self.query_one( Source ).show_file( self._args.file )
        self.query_one( DirectoryTree ).set_class( False, "visible" )
        self._init_tree()

### main.py ends here



"""Provides a widget for displaying information about a node from the AST tree."""

##############################################################################
# Textual imports.
from textual.app     import ComposeResult
from textual.widgets import Static

##############################################################################
# Local imports.
from .astview     import ASTNode
from .source_info import SourceInfo

##############################################################################
class NodeInfo( SourceInfo ):
    """Displays some information about the node."""

    DEFAULT_CSS = """
    NodeInfo {
        background: $panel;
        padding-left: 1;
        height: 1;
        border: none;
    }

    NodeInfo Static#--node-path {
        width: 1fr;
    }

    NodeInfo Static#--node-location {
        dock: right;
        width: 20;
    }
    """

    def compose( self ) -> ComposeResult:
        """Compose the main app screen.

        Returns:
            ComposeResult: The result of composing the screen.
        """
        yield Static( id="--node-path")
        yield Static( id="--node-location" )

    def show( self, node: ASTNode ) -> None:
        """Show some details for the given node.

        Args:
            node (ASTNode): The node to show the data for.
        """
        if ( loc := self.file_location_of( node ) ) is not None:
            self.query_one( "#--node-location", Static ).update(
                f"{loc[ 0 ]:04d}:{loc[ 1 ]:03d} -> {loc[ 2 ]:04d}:{loc[ 3 ]:03d}"
            )
        else:
            self.query_one( "#--node-location", Static ).update( "" )
        self.query_one( "#--node-path", Static ).update(
            " > ".join( reversed( list( self.path_from( node ) ) ) )
        )

### node_info.py ends here



"""Provides a widget for displaying the source related to the AST view."""

##############################################################################
# Python imports.
from pathlib   import Path
from typing    import Any, ClassVar, Final
from itertools import islice

##############################################################################
# Rich imports.
from rich.syntax import Syntax

##############################################################################
# Textual imports.
from textual          import events
from textual.app      import ComposeResult
from textual.widgets  import Static
from textual.geometry import Region
from textual.reactive import reactive

##############################################################################
# Local imports.
from .astview     import ASTNode
from .source_info import SourceInfo

##############################################################################
class Source( SourceInfo, can_focus=True ):
    """Displays the source code for the file."""

    COMPONENT_CLASSES: ClassVar[ set[ str ] ] = {
        "source--ast-node-dark-mode-highlight",
        "source--ast-node-dark-mode-highlight-1",
        "source--ast-node-dark-mode-highlight-2",
        "source--ast-node-dark-mode-highlight-3",
        "source--ast-node-dark-mode-highlight-4",
        "source--ast-node-dark-mode-highlight-5",
        "source--ast-node-light-mode-highlight",
        "source--ast-node-light-mode-highlight-1",
        "source--ast-node-light-mode-highlight-2",
        "source--ast-node-light-mode-highlight-3",
        "source--ast-node-light-mode-highlight-4",
        "source--ast-node-light-mode-highlight-5",
    }
    """set[ str ]: Classes that can be used to style the source view."""

    DEFAULT_CSS = """
    Source {
        height: 1fr;
        border: solid $primary-background-lighten-2;
        background: $panel;
    }

    Source:focus {
        border: double $primary-lighten-2;
    }

    Source > .source--ast-node-dark-mode-highlight   { background: #700; text-style: bold italic; }
    Source > .source--ast-node-dark-mode-highlight-1 { background: #606; }
    Source > .source--ast-node-dark-mode-highlight-2 { background: #055; }
    Source > .source--ast-node-dark-mode-highlight-3 { background: #440; }
    Source > .source--ast-node-dark-mode-highlight-4 { background: #303; }
    Source > .source--ast-node-dark-mode-highlight-5 { background: #333; }

    Source > .source--ast-node-light-mode-highlight   { background: #F00; text-style: bold italic; }
    Source > .source--ast-node-light-mode-highlight-1 { background: #E0E; }
    Source > .source--ast-node-light-mode-highlight-2 { background: #0DD; }
    Source > .source--ast-node-light-mode-highlight-3 { background: #CC0; }
    Source > .source--ast-node-light-mode-highlight-4 { background: #B0B; }
    Source > .source--ast-node-light-mode-highlight-5 { background: #AAA; }
    """

    MAX_ANCESTOR: Final = 5
    """int: The maximum number of ancestors that can receive rainbow highlights."""

    DEFAULT_DARK_THEME: Final = "github-dark"
    """str: The default dark theme to use for the source."""

    DEFAULT_LIGHT_THEME: Final = "xcode"
    """str: The default light theme to use for the source."""

    dark = reactive( True )
    """bool: Should the source be shown in a dark theme?"""

    def __init__(
            self, source: Path, *args: Any,
            dark_theme: str=DEFAULT_DARK_THEME,
            light_theme: str=DEFAULT_LIGHT_THEME,
            **kwargs: Any
    ) -> None:
        """Initialise the source viewing widget.

        Args:
            dark_theme (str, optional): The theme to use for dark mode.
            light_theme (str, optional)): The theme to use for light mode.
        """
        super().__init__( *args, **kwargs )
        self._source_file = source
        self._dark_theme  = dark_theme
        self._light_theme = light_theme

    def _build_source( self ) -> Syntax:
        """Builds a new `Stntax` of the file being viewed.

        Returns:
            Syntax: The resulting `Syntax` instance.

        Note:
            This updates `_source`, but also returns is as a kindness.
        """
        self._source = Syntax.from_path(
            str( self._source_file ), line_numbers=True, lexer="python",
            theme=self._dark_theme if self.dark else self._light_theme
        )
        return self._source

    def _populate_source( self ) -> None:
        """Repopulate the source display.

        Note:
            This should only be used when everything about the source has
            changed.
        """
        self.query_one( Static ).update( self._build_source() )

    def show_file( self, new_file: Path ) -> None:
        """Show a new file in the widget.

        Args:
            new_file (Path): The new file to show.
        """
        self._source_file = new_file
        self._populate_source()

    def compose( self ) -> ComposeResult:
        """Compose the source display.

        Returns:
            ComposeResult: The result of composing the source display.
        """
        yield Static()

    def on_mount( self ) -> None:
        """Configure the widget after the DOM is up and going."""
        self._populate_source()

    def watch_dark( self, _: bool ) -> None:
        """React to our own dark mode toggle.

        Note:
            This is only here while Textual's `watch` isn't behaving well.
        """
        self._populate_source()

    def on_click( self, _: events.Click ) -> None:
        """React to a mouse click."""
        self.focus()

    @property
    def _highlight_style( self ) -> str:
        return f"source--ast-node-{'dark' if self.dark else 'light'}-mode-highlight"

    def _highlight_ancestors( self, node: ASTNode ) -> None:
        """Apply highlighting to location-based ancestors of the given node.

        Args:
            node (ASTNode): The node to find the ancestors of.
        """
        for rule, ancestor in reversed( list( enumerate( islice( self.file_location_path_from( node ), 1, self.MAX_ANCESTOR + 1 ) ) ) ):
            self._source.stylize_range(
                self.get_component_rich_style( f"{self._highlight_style}-{rule + 1}", partial=True ), ancestor[ :2 ], ancestor[ 2: ]
            )

    def highlight( self, node: ASTNode, rainbow: bool=False ) -> None:
        """Highlight the given AST data in the source.

        Args:
            node (ASTNode): The AST node to highlight.
            rainbow (bool, optional): Use 'rainbow' highlighting?
        """

        # Sneaky nuking of any old styles. Rich's `Syntax` doesn't allow for
        # removing any stylized ranges so, other than nuking the while
        # Syntax and generating it again every time (slow!), the only option
        # we've got is to kill the list via an internal.
        #
        # pylint:disable=protected-access
        self._source._stylized_ranges = []

        # If rainbow highlighting is in effect, let's style the ancestors
        # first, so the current highlight is always at the top of the
        # styles.
        if rainbow:
            self._highlight_ancestors( node )

        # If the current node has a file location...
        if ( loc := self.file_location_of( node ) ):
            # ...highlight and scroll to it.
            self._source.stylize_range(
                self.get_component_rich_style( self._highlight_style, partial=True ), loc[ :2 ], loc[ 2: ]
            )
            line, _, end_line, _ = loc
            self.scroll_to_region( Region( y=line - 1, height=( end_line - line ) + 1 ) )

        # Update the display.
        self.query_one( Static ).refresh()

### source.py ends here



"""An AST viewer widget for Textual."""

##############################################################################
# Backward compatibility.
from __future__ import annotations

##############################################################################
# Python imports.
import ast
from pathlib   import Path
from functools import singledispatchmethod
from typing    import Any, cast, ClassVar, Union, Optional

##############################################################################
# Rich imports.
from rich.text import Text

##############################################################################
# Textual imports.
from textual.widgets import Tree, TreeNode
from textual.message import Message

##############################################################################
ASTNode = TreeNode[ Any ]

##############################################################################
class ASTView( Tree[ Any ] ):
    """The tree view of the AST."""

    # Note the hand-merged complement classes, keeping those of Tree.
    COMPONENT_CLASSES: ClassVar[ set[ str ] ] = Tree.COMPONENT_CLASSES | {
        "astview--def-name"
    }
    """set[ str ]: Classes that can be used to style the AST view."""

    DEFAULT_CSS = """
    ASTView {
        width: 1fr;
    }

    .astview--def-name {
        text-style: italic;
        color: #888;
    }
    """
    """str: The default CSS for the component."""

    def __init__( self, module: Path, *args: Any, name_defs: bool=True, **kwargs: Any ) -> None:
        """Initialise the view of the AST for the given module.

        Args:
            module (Path): The path of the module to show.
            name_defs (bool, optional): Include names in definition-type nodes?

        Note:
            By default function and class definition nodes will
            parenthetically show the name of the thing being defined. Use
            `name_defs` to turn this off.
        """

        # Remember some key stuff
        self._name_defs   = name_defs
        self._module_path = module

        # Get the AST parsed in. I *suppose* it would make sense to try and
        # do this in an async way, perhaps? But for now anyway let's just
        # parse in the whole blasted thing up front. I've never found the
        # AST parser to be noticeably slow.
        try:
            self._module: Optional[ ast.Module ] = ast.parse( module.read_text() )
        except SyntaxError:
            self._module = None

        # Now that we've configured things, get the Tree to do its own
        # thing.
        kwargs[ "label" ] = module.name
        kwargs[ "data" ]  = self._module_path
        super().__init__( *args, **kwargs )

    @property
    def module_path( self ) -> Path:
        """Path: The path for the module being shown."""
        return self._module_path

    @singledispatchmethod
    def base_node( self, item: Any, to_node: ASTNode ) -> ASTNode:
        """Attach a base node.

        Args:
            item (Any): The item to associate with the node.
            to_node (ASTNode): The node to attach to.

        Returns:
            ASTNode: The new node.
        """
        return to_node.add( item.__class__.__name__, data=item )

    @base_node.register
    def _( self, item: ast.AST, to_node: ASTNode ) -> ASTNode:
        """Attach a base node.

        Args:
            item (str): The item to associate with the node.
            to_node (ASTNode): The node to attach to.

        Returns:
            ASTNode: The new node.
        """
        # For most AST types the label will just be the class name.
        label = Text( item.__class__.__name__ )

        # But, for some, let's include some extra detail too to help people
        # visually scan the display.
        if self._name_defs and  isinstance( item, ( ast.ClassDef, ast.FunctionDef, ast.AsyncFunctionDef ) ):
            ( def_name := Text( f"({item.name})" ) ).stylize(
                self.get_component_rich_style( "astview--def-name", partial=True )
            )
            label = Text( f"{label} " ) + def_name

        # Now go with what we've got.
        return to_node.add( label, data=item )

    @base_node.register
    def _( self, item: str, to_node: ASTNode ) -> ASTNode:
        """Attach a base node.

        Args:
            item (str): The item to associate with the node.
            to_node (ASTNode): The node to attach to.

        Returns:
            ASTNode: The new node.
        """
        return to_node.add( item, data=item )

    @staticmethod
    def maybe_add( value: Any ) -> bool:
        """Does the value look like it should be added to the display?

        Args:
            value (Any): The value to consider.

        Returns:
            bool: `True` if the value should be added, `False` if not.
        """
        # If it's a sequence of some description, and it's empty, perhaps
        # not add it. Otherwise do add it. Note the over-protective casting;
        # mypy seems fine about the code without it (and actually moans when
        # I have it), pyright on the other hand wants me to be
        # super-specific here it seems.
        return (
            bool( cast( Union[ list[ Any ], tuple[ Any, ... ] ], value ) ) # type: ignore
            if isinstance( value, ( list, tuple ) ) else True
        )

    @singledispatchmethod
    def add( self, item: Any, to_node: ASTNode ) -> None:
        """Add an entry to the tree.

        Args:
            item (Any): The ast entry to add.
            to_node (ASTNode): The node to add to.
        """
        # If we've been given a list or a tuple...
        if isinstance( item, ( list, tuple ) ):
            # ...let's unroll what's inside, attaching everything to the
            # given node. (the over-protective cast is to keep pyright
            # happy; the ignore if it is to keep mypy happy).
            for row in cast( Union[ list[ Any ], tuple[ Any, ... ] ], item ): # type: ignore
                self.add( row, to_node )
        else:
            # Not a list or similar, so mark the item as a leaf and str it.
            to_node.add_leaf( repr( item ), data=item )

    @add.register
    def _( self, item: ast.AST, to_node: ASTNode ) -> None:
        """Add an AST item to the tree.

        Args:
            item (ast.AST): The ast entry to add.
            to_node (ASTNode): The node to add to.
        """

        # Create a new node to hand stuff off.
        node = self.base_node( item, to_node  )

        # If the AST item has any fields...
        if item._fields:
            # ...add them all.
            for field in item._fields:
                value = getattr( item, field )
                if self.maybe_add( value ):
                    self.add( value, self.base_node( field, node ) )
        else:
            # It's a AST item with no fields, so don't allow expanding as
            # there's nothing to expand.
            node.allow_expand = False

    def on_mount( self ) -> None:
        """Populate the view once we're mounted."""
        self.add( self._module, self.root )
        self.select_node( self.root )

    class NodeHighlighted( Message ):
        """Message sent when a node is highlighted in the tree view."""

        def __init__( self, sender: "ASTView", node: ASTNode ):
            """Initialise the node highlighted message."""
            self.node = node
            super().__init__( sender )

    def watch_cursor_line( self, previous_line: int, line: int ) -> None:
        """React to the cursor line changing.

        Args:
            previous_line (int): The old value.
            line (int): The new value.
        """
        super().watch_cursor_line( previous_line, line )
        if self.cursor_node is not None:
            self.emit_no_wait( self.NodeHighlighted( self, self.cursor_node ) )

### astview.py ends here



"""A base class for widgets that show information about the source."""

##############################################################################
# Backward compatibility.
from __future__ import annotations

##############################################################################
# Python imports.
import ast
from typing    import Iterator, Final
from pathlib   import Path
from itertools import groupby

##############################################################################
# Textual imports.
from textual.containers import Vertical

##############################################################################
# Local imports.
from .astview import ASTNode

##############################################################################
class SourceInfo( Vertical ):
    """Base class for information relating to showing source."""

    LOCATION_PROPS: Final = ( "lineno", "col_offset", "end_lineno", "end_col_offset" )
    """tuple[ str, ...]: The properties needed to highlight a location in the source."""

    @classmethod
    def has_location( cls, node: ASTNode ) -> bool:
        """Does the given node  have location information?

        Args:
            item (ASTNode): Am item from the AST.

        Returns:
            bool: `True` if the item has location information, otherwise `False.
        """
        return isinstance( node.data, ast.AST ) and all(
            hasattr( node.data, prop ) for prop in cls.LOCATION_PROPS
        )

    @classmethod
    def file_location_of( cls, node: ASTNode ) -> tuple[ int, int, int, int ] | None:
        """Find the file location of the given node.

        Args:
            node (ASTNode): The node in the tree to find the location of.

        Returns:
            tuple[ int, int, int, int ] | None: The location if one can be found.
        """
        # Textual's TreeNode doesn't currently allow access to the parent of
        # a node, so in here we're going to do some sneaky dipping into
        # internals. Hence:
        #
        # pylint:disable=protected-access
        if cls.has_location( node ):
            if ( item := node.data ) is not None: # Keep type checkers happy.
                return item.lineno, item.col_offset, item.end_lineno, item.end_col_offset
        elif node._parent is not None:
            return cls.file_location_of( node._parent )
        return None

    @classmethod
    def path_from( cls, node: ASTNode ) -> Iterator[ str ]:
        """Generate the path to the root from the given node.

        Args:
            node (ASTNode): The AST node to get the path from.

        Yields:
            str: The text for the node.
        """

        if isinstance( node.data, ast.AST ):
            yield node.data.__class__.__name__
        elif isinstance( node.data, Path ):
            yield node.data.name
        else:
            yield str( node.data )

        # Textual's TreeNode doesn't currently allow access to the parent of
        # a node, so in here we're going to do some sneaky dipping into
        # internals. Hence:
        #
        # pylint:disable=protected-access
        if node._parent is not None:
            yield from cls.path_from( node._parent )

    @classmethod
    def _file_location_path_from( cls, node: ASTNode ) -> Iterator[ tuple[ int, int, int, int ] ]:
        """Generate a file location path to the root from the given node.

        Args:
            node (ASTNode): The AST node to get the path from.

        Yields:
            tuple[ int, int, int, int ]: The file location.

        Note:
            This is a helper method for the non-internal of the same name.
        """

        # If this node has location data...
        if ( location := cls.file_location_of( node ) ) is not None:
            # ...make that available.
            yield location

        # Textual's TreeNode doesn't currently allow access to the parent of
        # a node, so in here we're going to do some sneaky dipping into
        # internals. Hence:
        #
        # pylint:disable=protected-access
        if node._parent is not None:
            yield from cls.file_location_path_from( node._parent )

    @classmethod
    def file_location_path_from( cls, node: ASTNode ) -> Iterator[ tuple[ int, int, int, int ] ]:
        """Generate a file location path to the root from the given node.

        Args:
            node (ASTNode): The AST node to get the path from.

        Yields:
            tuple[ int, int, int, int ]: The file location.
        """
        yield from ( pos for pos, _ in groupby( cls._file_location_path_from( node ) ) )

### source_info.py ends here
